# -*- coding: utf-8 -*-

import unittest

import pytest
import yaml
import pandas as pd
import numpy as np

from gordo.machine.model.transformers.imputer import InfImputer
from gordo import serializer

from sklearn.preprocessing import FunctionTransformer
from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA


class GordoFunctionTransformerFuncsTestCase(unittest.TestCase):
    """
    Test all functions within gordo meants for use in a Scikit-Learn
    FunctionTransformer work as expected
    """

    def _validate_transformer(self, transformer):
        """
        Inserts a transformer into the middle of a pipeline and runs it
        """
        pipe = Pipeline([("pca1", PCA()), ("custom", transformer), ("pca2", PCA())])
        X = np.random.random(size=100).reshape(10, 10)
        pipe.fit_transform(X)

    def test_multiply_by_function_transformer(self):
        from gordo.machine.model.transformer_funcs.general import multiply_by

        # Provide a require argument
        tf = FunctionTransformer(func=multiply_by, kw_args={"factor": 2})
        self._validate_transformer(tf)

        # Ignore the required argument
        tf = FunctionTransformer(func=multiply_by)
        with self.assertRaises(TypeError):
            self._validate_transformer(tf)


@pytest.mark.parametrize("strategy", ["extremes", "minmax"])
def test_infimputer_basic(strategy):
    """
    Functionality of the InfImputer
    """
    base_x = np.random.random((100, 10)).astype(np.float32)

    flat_view = base_x.ravel()

    pos_inf_idxs = np.random.randint(0, len(flat_view), size=100)
    neg_inf_idxs = np.random.randint(0, len(flat_view), size=100)

    flat_view[pos_inf_idxs] = np.inf
    flat_view[neg_inf_idxs] = -np.inf

    # Our base x should now be littered with pos/neg inf values
    assert np.isposinf(base_x).sum() > 0, "Expected some positive infinity values here"
    assert np.isneginf(base_x).sum() > 0, "Expected some negative infinity values here"

    imputer = InfImputer(strategy=strategy, delta=2.0)

    # Test imputer on numpy array
    X = base_x.copy()
    X = imputer.fit_transform(X)
    assert np.isposinf(X).sum() == 0, "Expected no positive infinity values here"
    assert np.isneginf(X).sum() == 0, "Expected no negative infinity values here"

    if strategy == "extremes":
        # All pos infs in base_x should be filled in X now as the max dtype value
        assert np.all(X[np.where(np.isposinf(base_x))] == np.finfo(X.dtype).max)

        # All neg infs in base_x should be filled in X now as the min dtype value
        assert np.all(X[np.where(np.isneginf(base_x))] == np.finfo(X.dtype).min)

    # min max is a bit more difficult to assert however.
    elif strategy == "minmax":

        # Identify the features in the base array which have pos infs
        features_with_pos_infs = np.where(np.isposinf(base_x))[1]

        # Get the maxes of those features in the imputed X
        filled_maxes = X[:, features_with_pos_infs].max(axis=0)

        # Get the previous maxes
        previous_maxes = np.ma.masked_invalid(base_x[:, features_with_pos_infs]).max(
            axis=0
        )

        # Compare that each new max is the previous max + 2.
        # These were previously 'inf' values, replaced by previous max + 2
        for filled_max, previous_max in zip(filled_maxes, previous_maxes):
            assert np.isclose(filled_max, previous_max + 2.0)

        ### Repeat the process for negative infs ###

        # Identify the features in the base array which have pos infs
        features_with_neg_infs = np.where(np.isneginf(base_x))[1]

        # Get the maxes of those features in the imputed X
        filled_mins = X[:, features_with_neg_infs].min(axis=0)

        # Get the previous maxes
        previous_mins = np.ma.masked_invalid(base_x[:, features_with_neg_infs]).min(
            axis=0
        )

        # Now compare, that each new min is the min - 2. wq
        for filled_min, previous_min in zip(filled_mins, previous_mins):
            assert np.isclose(filled_min, previous_min - 2.0)

    # Test imputer on pandas dataframe
    X = pd.DataFrame(base_x.copy())
    X = imputer.fit_transform(X)
    assert np.isposinf(X).sum() == 0, "Expected no positive infinity values here"
    assert np.isneginf(X).sum() == 0, "Expected no negative infinity values here"


def test_infimputer_fill_values():
    """
    InfImputer when fill values are provided
    """
    base_x = np.random.random((100, 10)).astype(np.float32)

    flat_view = base_x.ravel()

    pos_inf_idxs = [1, 2, 3, 4, 5]
    neg_inf_idxs = [6, 7, 8, 9, 10]

    flat_view[pos_inf_idxs] = np.inf
    flat_view[neg_inf_idxs] = -np.inf

    # Our base x should now be littered with pos/neg inf values
    assert np.isposinf(base_x).sum() > 0, "Expected some positive infinity values here"
    assert np.isneginf(base_x).sum() > 0, "Expected some negative infinity values here"

    imputer = InfImputer(inf_fill_value=9999.0, neg_inf_fill_value=-9999.0)
    X = imputer.fit_transform(base_x)
    np.equal(
        X.ravel()[[pos_inf_idxs]], np.array([9999.0, 9999.0, 9999.0, 9999.0, 9999.0])
    )
    np.equal(
        X.ravel()[[neg_inf_idxs]],
        np.array([-9999.0, -9999.0, -9999.0, -9999.0, -9999.0]),
    )


@pytest.mark.parametrize(
    "config_str",
    [
        """
    sklearn.pipeline.Pipeline:
      steps:
        - gordo.machine.model.transformers.imputer.InfImputer
    """,
        """
    sklearn.pipeline.Pipeline:
      steps:
        - gordo.machine.model.transformers.imputer.InfImputer:
            inf_fill_value: 10
    """,
        """gordo.machine.model.transformers.imputer.InfImputer""",
    ],
)
def test_imputer_from_definition(config_str: str):
    """
    Ensure it plays well with the gordo serializer
    """
    config = yaml.safe_load(config_str)
    model = serializer.from_definition(config)

    if isinstance(model, Pipeline):
        assert isinstance(model.steps[-1][1], InfImputer)
    else:
        assert isinstance(model, InfImputer)

    serializer.from_definition(serializer.into_definition(model))



import os
import pandas as pd

from pytest import fixture

from gordo.machine.dataset.data_provider.file_type import (
    CsvFileType,
    ParquetFileType,
    TimeSeriesColumns,
)


@fixture
def data_dir():
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), "data", "datalake")


def compare_dtype_names(dtypes, column_types):
    for column, type_name in column_types:
        try:
            dtype = dtypes[column]
        except KeyError:
            raise AssertionError(
                "Can not find '%s' column in pandas.DataFrame" % column
            )
        assert dtype.name == type_name, "Wrong dtype for column '%s'" % column


def test_csv_file_type(data_dir):
    csv_file = os.path.join(data_dir, "TRC-322", "TRC-322_2000.csv")
    time_series_columns = TimeSeriesColumns("Time", "Value", "Status")
    header = ["Sensor", "Value", "Time", "Status"]
    file_type = CsvFileType(header, time_series_columns)
    with open(csv_file, "rb") as f:
        df = file_type.read_df(f)
        assert len(df) == 10
        compare_dtype_names(df.dtypes, (("Value", "float32"), ("Status", "int64")))
        assert isinstance(df.index, pd.DatetimeIndex)


def test_parquet_file_type(data_dir):
    parquet_file = os.path.join(data_dir, "TRC-323", "parquet", "TRC-323_2001.parquet")
    time_series_columns = TimeSeriesColumns("Time", "Value", "Status")
    file_type = ParquetFileType(time_series_columns)
    with open(parquet_file, "rb") as f:
        df = file_type.read_df(f)
        assert len(df) == 20
        compare_dtype_names(df.dtypes, (("Value", "float64"), ("Status", "int32")))
        assert isinstance(df.index, pd.DatetimeIndex)



# -*- coding: utf-8 -*-

import ipaddress
import os
import typing

import yaml
import click
from dateutil import parser

from gordo.machine.dataset.data_provider import providers


class DataProviderParam(click.ParamType):
    """
    Load a DataProvider from JSON/YAML representation or from a JSON/YAML file
    """

    name = "data-provider"

    def convert(self, value, param, ctx):
        if os.path.isfile(value):
            with open(value) as f:
                kwargs = yaml.safe_load(f)
        else:
            kwargs = yaml.safe_load(value)

        if "type" not in kwargs:
            self.fail(f"Cannot create DataProvider without 'type' key defined")

        kind = kwargs.pop("type")

        Provider = getattr(providers, kind, None)
        if Provider is None:
            self.fail(f"No DataProvider named '{kind}'")
        return Provider(**kwargs)


class IsoFormatDateTime(click.ParamType):
    """
    Parse a string into an ISO formatted datetime object
    """

    name = "iso-datetime"

    def convert(self, value, param, ctx):
        try:
            return parser.isoparse(value)
        except ValueError:
            self.fail(f"Failed to parse date '{value}' as ISO formatted date'")


class HostIP(click.ParamType):
    """
    Validate input is a valid IP address
    """

    name = "host"

    def convert(self, value, param, ctx):
        try:
            ipaddress.ip_address(value)
            return value
        except ValueError as e:
            self.fail(e)


def key_value_par(val) -> typing.Tuple[str, str]:
    """
    Helpder for CLI input of 'key,val'
    """
    return val.split(",")



# -*- coding: utf-8 -*-

import abc
from typing import Optional, Union

import numpy as np
import pandas as pd


class GordoBase(abc.ABC):
    @abc.abstractmethod
    def __init__(self, **kwargs):
        """Initialize the model"""
        ...

    @abc.abstractmethod
    def get_params(self, deep=False):
        """Return a dict containing all parameters used to initialized object"""
        ...

    @abc.abstractmethod
    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ):
        """
        Score the model; must implement the correct default scorer based on model type
        """
        ...

    @abc.abstractmethod
    def get_metadata(self):
        """Get model specific metadata, if any"""



# -*- coding: utf-8 -*-

import logging

import numpy as np

from sklearn.pipeline import Pipeline

"""
Collection of functions to work with the input and output of a scikit.base.BaseEstimator
"""

logger = logging.getLogger(__name__)


def get_model_output(model: Pipeline, X: np.ndarray) -> np.ndarray:
    """
    Get the raw output from the current model given X.
    Will try to `predict` and then `transform`, raising an error
    if both fail.

    Parameters
    ----------
    X: np.ndarray
        2d array of sample(s)

    Returns
    -------
    np.ndarray
        The raw output of the model in numpy array form.
    """
    try:
        return model.predict(X)  # type: ignore

    # Model may only be a transformer
    except AttributeError:
        try:
            return model.transform(X)  # type: ignore
        except Exception as exc:
            logger.error(f"Failed to predict or transform; error: {exc}")
            raise



import os
import dateutil.parser
import logging

import pytest
import adal

from gordo.machine.dataset.data_provider.providers import DataLakeProvider
from gordo.machine.dataset import dataset
from gordo.machine.dataset.sensor_tag import normalize_sensor_tags


def _get_default_dataset_config():
    train_start_date = dateutil.parser.isoparse("2017-01-01T08:56:00+00:00")
    train_end_date = dateutil.parser.isoparse("2017-01-01T10:01:00+00:00")
    return {
        "type": "TimeSeriesDataset",
        "train_start_date": train_start_date,
        "train_end_date": train_end_date,
        "tag_list": normalize_sensor_tags(["TRC-FIQ -39-0706", "GRA-EM-23-0003ARV.PV"]),
        "data_provider": DataLakeProvider(),
    }


def test_get_data_serviceauth_fail(caplog):
    train_start_date = dateutil.parser.isoparse("2017-01-01T08:56:00+00:00")
    train_end_date = dateutil.parser.isoparse("2017-01-01T10:01:00+00:00")

    dataset_config = _get_default_dataset_config()
    dataset_config["train_start_date"] = train_start_date
    dataset_config["train_end_date"] = train_end_date
    dataset_config["data_provider"] = DataLakeProvider(
        dl_service_auth_str="TENTANT_UNKNOWN:BOGUS:PASSWORD"
    )

    dl_backed = dataset._get_dataset(dataset_config)

    with pytest.raises(adal.adal_error.AdalError), caplog.at_level(logging.CRITICAL):
        dl_backed.get_data()


def test_init():
    config = _get_default_dataset_config()
    dl_backed = dataset._get_dataset(config)
    assert (
        dl_backed is not None
    ), f"Failed to create dataset object of type {config['type']}"


@pytest.mark.skipif(
    os.getenv("INTERACTIVE") is None,
    reason="Skipping test, INTERACTIVE not set in environment variable",
)
def test_get_data_interactive():
    dataset_config = _get_default_dataset_config()
    dataset_config["data_provider"] = DataLakeProvider(interactive=True)
    dl_backed = dataset._get_dataset(dataset_config)
    data = dl_backed.get_data()
    assert len(data) >= 0


@pytest.mark.skipif(
    os.getenv("TEST_SERVICE_AUTH") is None,
    reason="Skipping test, TEST_SERVICE_AUTH not set in environment variable",
)
def test_get_data_serviceauth_in_config():
    dataset_config = _get_default_dataset_config()
    dataset_config["data_provider"] = DataLakeProvider(
        dl_service_auth_str=os.getenv("TEST_SERVICE_AUTH")
    )
    dataset_config["resolution"] = "10T"
    dl_backed = dataset._get_dataset(dataset_config)
    data, _ = dl_backed.get_data()

    assert dataset_config["tag_list"] == list(data.columns.values)

    expected_rows = 7
    assert (
        len(data) == expected_rows
    ), f"Default resolution 10 minutes should give {expected_rows} rows"

    assert (
        not data.isnull().values.any()
    ), "Resulting dataframe should not have any NaNs"



import functools
import inspect
from typing import Callable


def capture_args(method: Callable):
    """
    Decorator that captures args and kwargs passed to a given method.
    This assumes the decorated method has a self, which has a dict of
    kwargs assigned as an attribute named _params.

    Parameters
    ----------
    method: Callable
        Some method of an object, with 'self' as the first parameter.

    Returns
    -------
    Any
        Returns whatever the original method would return
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):

        sig_params = inspect.signature(method).parameters.items()

        # Get the default values for the method signature
        params = {
            param: value.default
            for param, value in sig_params
            if value.default is not inspect.Parameter.empty and param != "self"
        }

        # Update with arg values provided
        arg_map = dict()
        for arg_val, arg_key in zip(
            args, (arg for arg in inspect.getfullargspec(method).args if arg != "self")
        ):
            arg_map[arg_key] = arg_val

        # Update params with args/kwargs provided in the current call
        params.update(arg_map)
        params.update(kwargs)

        self._params = params
        return method(self, *args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-

from typing import List, Iterable, Optional

import pytest
import numpy as np
import pandas as pd
import dateutil.parser
from datetime import datetime

from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider
from gordo.machine.dataset.datasets import (
    RandomDataset,
    TimeSeriesDataset,
    InsufficientDataAfterRowFilteringError,
)
from gordo.machine.dataset.base import GordoBaseDataset, InsufficientDataError
from gordo.machine.dataset.sensor_tag import SensorTag


@pytest.fixture
def dataset():
    return RandomDataset(
        train_start_date="2017-12-25 06:00:00Z",
        train_end_date="2017-12-29 06:00:00Z",
        tag_list=[SensorTag("Tag 1", None), SensorTag("Tag 2", None)],
    )


def create_timeseries_list():
    """Create three dataframes with different resolution and different start/ends"""
    # Test for no NaNs, test for correct first and last date
    latest_start = "2018-01-03 06:00:00Z"
    earliest_end = "2018-01-05 06:00:00Z"

    index_seconds = pd.date_range(
        start="2018-01-01 06:00:00Z", end="2018-01-07 06:00:00Z", freq="S"
    )
    index_minutes = pd.date_range(
        start="2017-12-28 06:00:00Z", end=earliest_end, freq="T"
    )
    index_hours = pd.date_range(
        start=latest_start, end="2018-01-12 06:00:00Z", freq="H"
    )

    timeseries_seconds = pd.Series(
        data=np.random.randint(0, 100, len(index_seconds)),
        index=index_seconds,
        name="ts-seconds",
    )
    timeseries_minutes = pd.Series(
        data=np.random.randint(0, 100, len(index_minutes)),
        index=index_minutes,
        name="ts-minutes",
    )
    timeseries_hours = pd.Series(
        data=np.random.randint(0, 100, len(index_hours)),
        index=index_hours,
        name="ts-hours",
    )

    return (
        [timeseries_seconds, timeseries_minutes, timeseries_hours],
        latest_start,
        earliest_end,
    )


def test_random_dataset_attrs(dataset):
    """
    Test expected attributes
    """

    assert isinstance(dataset, GordoBaseDataset)
    assert hasattr(dataset, "get_data")
    assert hasattr(dataset, "get_metadata")

    X, y = dataset.get_data()
    assert isinstance(X, pd.DataFrame)

    # y can either be None or an numpy array
    assert isinstance(y, pd.DataFrame) or y is None

    metadata = dataset.get_metadata()
    assert isinstance(metadata, dict)


def test_join_timeseries(dataset):

    timeseries_list, latest_start, earliest_end = create_timeseries_list()

    assert len(timeseries_list[0]) > len(timeseries_list[1]) > len(timeseries_list[2])

    frequency = "7T"
    timedelta = pd.Timedelta("7 minutes")
    resampling_start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
    resampling_end = dateutil.parser.isoparse("2018-01-15 08:00:00Z")
    all_in_frame = dataset.join_timeseries(
        timeseries_list, resampling_start, resampling_end, frequency
    )

    # Check that first resulting resampled, joined row is within "frequency" from
    # the real first data point
    assert all_in_frame.index[0] >= pd.Timestamp(latest_start) - timedelta
    assert all_in_frame.index[-1] <= pd.Timestamp(resampling_end)


@pytest.mark.parametrize(
    "value,n_rows,resolution,row_threshold,error",
    [
        # Frequency passed as zero, resulting in an ZeroDivisionError during aggregation
        (None, None, "0T", 0, ZeroDivisionError),
        # Empty series results in an InsufficientDataError
        (None, 0, "12T", 0, InsufficientDataError),
        # When all rows are NaNs and dropped result in InsufficientDataError
        (np.NaN, None, "12T", 0, InsufficientDataError),
        # Rows less then or equal to `row_threshold` result in InsufficientDataError
        (None, 6, "12T", 6, InsufficientDataError),
    ],
)
def test_join_timeseries_empty_series(value, n_rows, resolution, row_threshold, error):
    """
    Test that empty data scenarios raise appropriate errors
    """
    train_start_date = dateutil.parser.isoparse("2018-01-01 00:00:00+00:00")
    train_end_date = dateutil.parser.isoparse("2018-01-05 00:00:00+00:00")
    tag_list = [SensorTag(name=n, asset=None) for n in ["Tag 1", "Tag 2", "Tag 3"]]

    kwargs = {
        "train_start_date": train_start_date,
        "train_end_date": train_end_date,
        "tag_list": tag_list,
        "resolution": resolution,
        "row_threshold": row_threshold,
        "data_provider": MockDataProvider(value=np.NaN, n_rows=n_rows),
    }

    with pytest.raises(error):
        TimeSeriesDataset(**kwargs).get_data()


def test_join_timeseries_nonutcstart(dataset):
    timeseries_list, latest_start, earliest_end = create_timeseries_list()
    frequency = "7T"
    resampling_start = dateutil.parser.isoparse("2017-12-25 06:00:00+07:00")
    resampling_end = dateutil.parser.isoparse("2018-01-12 13:07:00+07:00")
    all_in_frame = dataset.join_timeseries(
        timeseries_list, resampling_start, resampling_end, frequency
    )
    assert len(all_in_frame) == 1854


def test_join_timeseries_with_gaps(dataset):

    timeseries_list, latest_start, earliest_end = create_timeseries_list()

    assert len(timeseries_list[0]) > len(timeseries_list[1]) > len(timeseries_list[2])

    remove_from = "2018-01-03 10:00:00Z"
    remove_to = "2018-01-03 18:00:00Z"
    timeseries_with_holes = [
        ts[(ts.index < remove_from) | (ts.index >= remove_to)] for ts in timeseries_list
    ]

    frequency = "10T"
    resampling_start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
    resampling_end = dateutil.parser.isoparse("2018-01-12 07:00:00Z")

    all_in_frame = dataset.join_timeseries(
        timeseries_with_holes, resampling_start, resampling_end, frequency
    )
    assert all_in_frame.index[0] == pd.Timestamp(latest_start)
    assert all_in_frame.index[-1] == pd.Timestamp(resampling_end)

    expected_index = pd.date_range(
        start=dateutil.parser.isoparse(latest_start), end=resampling_end, freq=frequency
    )
    assert list(all_in_frame.index) == list(expected_index)


def test_row_filter():
    """Tests that row_filter filters away rows"""
    kwargs = dict(
        data_provider=MockDataProvider(),
        tag_list=[
            SensorTag("Tag 1", None),
            SensorTag("Tag 2", None),
            SensorTag("Tag 3", None),
        ],
        train_start_date=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
        train_end_date=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
    )
    X, _ = TimeSeriesDataset(**kwargs).get_data()
    assert 577 == len(X)

    X, _ = TimeSeriesDataset(row_filter="`Tag 1` < 5000", **kwargs).get_data()
    assert 8 == len(X)

    X, _ = TimeSeriesDataset(
        row_filter="`Tag 1` / `Tag 3` < 0.999", **kwargs
    ).get_data()
    assert 3 == len(X)


def test_aggregation_methods():
    """Tests that it works to set aggregation method(s)"""

    kwargs = dict(
        data_provider=MockDataProvider(),
        tag_list=[
            SensorTag("Tag 1", None),
            SensorTag("Tag 2", None),
            SensorTag("Tag 3", None),
        ],
        train_start_date=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
        train_end_date=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
    )

    # Default aggregation gives no extra columns
    X, _ = TimeSeriesDataset(**kwargs).get_data()
    assert (577, 3) == X.shape

    # The default single aggregation method gives the tag-names as columns
    assert list(X.columns) == ["Tag 1", "Tag 2", "Tag 3"]

    # Using two aggregation methods give a multi-level column with tag-names
    # on top and aggregation_method as second level
    X, _ = TimeSeriesDataset(aggregation_methods=["mean", "max"], **kwargs).get_data()

    assert (577, 6) == X.shape
    assert list(X.columns) == [
        ("Tag 1", "mean"),
        ("Tag 1", "max"),
        ("Tag 2", "mean"),
        ("Tag 2", "max"),
        ("Tag 3", "mean"),
        ("Tag 3", "max"),
    ]


def test_time_series_no_resolution():
    kwargs = dict(
        data_provider=MockDataProvider(),
        tag_list=[
            SensorTag("Tag 1", None),
            SensorTag("Tag 2", None),
            SensorTag("Tag 3", None),
        ],
        train_start_date=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
        train_end_date=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
    )

    no_resolution, _ = TimeSeriesDataset(resolution=None, **kwargs).get_data()
    wi_resolution, _ = TimeSeriesDataset(resolution="10T", **kwargs).get_data()
    assert len(no_resolution) > len(wi_resolution)


@pytest.mark.parametrize(
    "tag_list",
    [
        [SensorTag("Tag 1", None), SensorTag("Tag 2", None), SensorTag("Tag 3", None)],
        [SensorTag("Tag 1", None)],
    ],
)
@pytest.mark.parametrize(
    "target_tag_list",
    [
        [SensorTag("Tag 2", None), SensorTag("Tag 1", None), SensorTag("Tag 3", None)],
        [SensorTag("Tag 1", None)],
        [SensorTag("Tag10", None)],
        [],
    ],
)
def test_timeseries_target_tags(tag_list, target_tag_list):
    start = dateutil.parser.isoparse("2017-12-25 06:00:00Z")
    end = dateutil.parser.isoparse("2017-12-29 06:00:00Z")
    tsd = TimeSeriesDataset(
        start,
        end,
        tag_list=tag_list,
        target_tag_list=target_tag_list,
        data_provider=MockDataProvider(),
    )
    X, y = tsd.get_data()

    assert len(X) == len(y)

    # If target_tag_list is empty, it defaults to tag_list
    if target_tag_list:
        assert y.shape[1] == len(target_tag_list)
    else:
        assert y.shape[1] == len(tag_list)

    # Ensure the order in maintained
    assert [tag.name for tag in target_tag_list or tag_list] == y.columns.tolist()

    # Features should match the tag_list
    assert X.shape[1] == len(tag_list)

    # Ensure the order in maintained
    assert [tag.name for tag in tag_list] == X.columns.tolist()


class MockDataProvider(GordoBaseDataProvider):
    def __init__(self, value=None, n_rows=None, **kwargs):
        """With value argument for generating different types of data series (e.g. NaN)"""
        self.value = value
        self.n_rows = n_rows

    def can_handle_tag(self, tag):
        return True

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[SensorTag],
        dry_run: Optional[bool] = False,
    ) -> Iterable[pd.Series]:
        index = pd.date_range(train_start_date, train_end_date, freq="s")
        for i, name in enumerate(sorted([tag.name for tag in tag_list])):
            # If value not passed, data for each tag are staggered integer ranges
            data = [self.value if self.value else i for i in range(i, len(index) + i)]
            series = pd.Series(index=index, data=data, name=name)
            yield series[: self.n_rows] if self.n_rows else series


def test_timeseries_dataset_compat():
    """
    There are accepted keywords in the config file when using type: TimeSeriesDataset
    which don't actually match the kwargs of the dataset's __init__; for compatibility
    :func:`gordo.machine.dataset.datasets.compat` should adjust for these differences.
    """
    dataset = TimeSeriesDataset(
        data_provider=MockDataProvider(),
        train_start_date="2017-12-25 06:00:00Z",
        train_end_date="2017-12-29 06:00:00Z",
        tags=[SensorTag("Tag 1", None)],
    )
    assert dataset.train_start_date == dateutil.parser.isoparse("2017-12-25 06:00:00Z")
    assert dataset.train_end_date == dateutil.parser.isoparse("2017-12-29 06:00:00Z")
    assert dataset.tag_list == [SensorTag("Tag 1", None)]


@pytest.mark.parametrize("n_samples_threshold, filter_value", [(10, 5000), (0, 100)])
def test_insufficient_data_after_row_filtering(n_samples_threshold, filter_value):
    """
    Test that dataframe after row_filter scenarios raise appropriate
    InsufficientDataAfterRowFilteringError
    """

    kwargs = dict(
        data_provider=MockDataProvider(),
        tag_list=[
            SensorTag("Tag 1", None),
            SensorTag("Tag 2", None),
            SensorTag("Tag 3", None),
        ],
        train_start_date=dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
        train_end_date=dateutil.parser.isoparse("2017-12-29 06:00:00Z"),
        n_samples_threshold=n_samples_threshold,
    )

    with pytest.raises(InsufficientDataAfterRowFilteringError):
        TimeSeriesDataset(row_filter=f"`Tag 1` < {filter_value}", **kwargs).get_data()



# -*- coding: utf-8 -*-

import ast
import json
import logging
from io import StringIO
from datetime import datetime, timezone

import pytest
import yaml

from gordo import __version__
from gordo.machine.dataset.datasets import TimeSeriesDataset
from gordo.machine import Machine
from gordo.workflow.config_elements.normalized_config import NormalizedConfig
from gordo.workflow.workflow_generator.workflow_generator import get_dict_from_yaml


logger = logging.getLogger(__name__)


def test_dataset_from_dict():
    """
    Test ability to create a Dataset from a config element
    """
    element_str = """
        name: ct-23-0002
        dataset:
          resolution: 2T
          tag_list:
            - GRA-YE  -23-0751X.PV
            - GRA-TE  -23-0698.PV
            - GRA-PIT -23-0619B.PV
          train_start_date: 2011-05-20T01:00:04+02:00
          train_end_date: 2018-05-10T15:05:50+02:00
    """
    dataset_config = get_dict_from_yaml(StringIO(element_str))["dataset"]
    dataset = TimeSeriesDataset.from_dict(dataset_config.copy())
    asdict = dataset.to_dict()
    assert asdict["tag_list"] == [
        "GRA-YE  -23-0751X.PV",
        "GRA-TE  -23-0698.PV",
        "GRA-PIT -23-0619B.PV",
    ]
    assert asdict["resolution"] == "2T"
    assert asdict["train_start_date"] == "2011-05-20T01:00:04+02:00"
    assert asdict["train_end_date"] == "2018-05-10T15:05:50+02:00"


def test_dataset_from_config_checks_dates():
    """
    A dataset needs to have train_start_date properly before train_end_date
    """
    element_str = """
        dataset:
          resolution: 2T
          tags:
            - GRA-YE  -23-0751X.PV
            - GRA-TE  -23-0698.PV
            - GRA-PIT -23-0619B.PV
          train_start_date: 2018-05-10T15:05:50+02:00
          train_end_date: 2018-05-10T15:05:50+02:00
    """
    dataset_config = yaml.load(element_str, Loader=yaml.FullLoader)["dataset"]
    with pytest.raises(ValueError):
        TimeSeriesDataset.from_dict(dataset_config)


@pytest.fixture
def default_globals():
    default_globals = dict(NormalizedConfig.DEFAULT_CONFIG_GLOBALS)
    # set something different here so we dont have to change the test every time we
    # change the default runtime parameters
    default_globals["runtime"] = {
        "server": {
            "resources": {
                "requests": {"memory": 1, "cpu": 2},
                "limits": {"memory": 3, "cpu": 4},
            }
        }
    }
    default_globals["dataset"] = {"asset": "global-asset"}
    return default_globals


def test_machine_from_config(default_globals: dict):
    """
    Test ability to create a Machine from a config element.
    """

    element_str = """
        name: ct-23-0001-machine
        data_provider:
          threads: 10
        dataset:
          tags: [GRA-TE  -23-0733.PV, GRA-TT  -23-0719.PV, GRA-YE  -23-0751X.PV]
          target_tag_list: [GRA-TE -123-456]
          train_start_date: 2018-01-01T09:00:30Z
          train_end_date: 2018-01-02T09:00:30Z
        model:
          sklearn.pipeline.Pipeline:
            steps:
              - sklearn.preprocessing.data.MinMaxScaler
              - gordo.machine.model.models.KerasAutoEncoder:
                  kind: feedforward_hourglass
        evaluation:
            scoring_scaler: Null
        metadata:
          id: special-id
    """
    element = get_dict_from_yaml(StringIO(element_str))
    machine = Machine.from_config(
        element, project_name="test-project-name", config_globals=default_globals
    )
    logger.info(f"{machine}")
    assert isinstance(machine, Machine)
    assert len(machine.dataset.tag_list) == 3

    # The metadata of machine should be json serializable
    json.dumps(machine.to_dict()["metadata"])

    # The metadata of machine should be ast.literal_eval-able when cast as a str
    assert (
        ast.literal_eval(str(machine.to_dict()["metadata"]))
        == machine.to_dict()["metadata"]
    )
    # dictionary representation of the machine expected:
    expected = {
        "dataset": {
            "aggregation_methods": "mean",
            "asset": "global-asset",
            "data_provider": {
                "dl_service_auth_str": None,
                "interactive": False,
                "storename": "dataplatformdlsprod",
                "type": "DataLakeProvider",
            },
            "default_asset": None,
            "n_samples_threshold": 0,
            "resolution": "10T",
            "row_filter": "",
            "row_filter_buffer_size": 0,
            "tag_list": [
                "GRA-TE  -23-0733.PV",
                "GRA-TT  -23-0719.PV",
                "GRA-YE  -23-0751X.PV",
            ],
            "target_tag_list": ["GRA-TE -123-456"],
            "train_end_date": "2018-01-02T09:00:30+00:00",
            "train_start_date": "2018-01-01T09:00:30+00:00",
            "type": "TimeSeriesDataset",
        },
        "evaluation": {
            "cv_mode": "full_build",
            "metrics": [
                "explained_variance_score",
                "r2_score",
                "mean_squared_error",
                "mean_absolute_error",
            ],
            "scoring_scaler": None,
        },
        "metadata": {
            "build_metadata": {
                "model": {
                    "cross_validation": {
                        "cv_duration_sec": None,
                        "scores": {},
                        "splits": {},
                    },
                    "model_builder_version": __version__,
                    "model_creation_date": None,
                    "model_meta": {},
                    "model_offset": 0,
                    "model_training_duration_sec": None,
                },
                "dataset": {"query_duration_sec": None, "dataset_meta": {}},
            },
            "user_defined": {
                "global-metadata": {},
                "machine-metadata": {"id": "special-id"},
            },
        },
        "model": {
            "sklearn.pipeline.Pipeline": {
                "steps": [
                    "sklearn.preprocessing.data.MinMaxScaler",
                    {
                        "gordo.machine.model.models.KerasAutoEncoder": {
                            "kind": "feedforward_hourglass"
                        }
                    },
                ]
            }
        },
        "name": "ct-23-0001-machine",
        "project_name": "test-project-name",
        "runtime": {
            "reporters": [],
            "server": {
                "resources": {
                    "limits": {"cpu": 4, "memory": 3},
                    "requests": {"cpu": 2, "memory": 1},
                }
            },
        },
    }
    assert machine.to_dict() == expected


def test_invalid_model(default_globals: dict):
    """
    Test invalid model with 'step' instead of 'steps'
    """
    element_str = """
        name: ct-23-0001-machine
        data_provider:
          threads: 10
        dataset:
          tags: [GRA-TE  -23-0733.PV, GRA-TT  -23-0719.PV, GRA-YE  -23-0751X.PV]
          target_tag_list: [GRA-TE -123-456]
          train_start_date: 2018-01-01T09:00:30Z
          train_end_date: 2018-01-02T09:00:30Z
        model:
          sklearn.pipeline.Pipeline:
            step:
              - sklearn.preprocessing.data.MinMaxScaler
              - gordo.machine.model.models.KerasAutoEncoder:
                  kind: feedforward_hourglass
        evaluation:
            scoring_scaler: Null
        metadata:
          id: special-id
    """
    element = get_dict_from_yaml(StringIO(element_str))
    with pytest.raises(ValueError):
        Machine.from_config(
            element, project_name="test-project-name", config_globals=default_globals
        )



# -*- coding: utf-8 -*-

from typing import Tuple, Dict, Any, Union

from tensorflow.keras.optimizers import Optimizer
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Dense
from tensorflow import keras

from tensorflow.keras.models import Sequential as KerasSequential
from gordo.machine.model.register import register_model_builder
from gordo.machine.model.factories.utils import hourglass_calc_dims, check_dim_func_len


@register_model_builder(type="KerasAutoEncoder")
def feedforward_model(
    n_features: int,
    n_features_out: int = None,
    encoding_dim: Tuple[int, ...] = (256, 128, 64),
    encoding_func: Tuple[str, ...] = ("tanh", "tanh", "tanh"),
    decoding_dim: Tuple[int, ...] = (64, 128, 256),
    decoding_func: Tuple[str, ...] = ("tanh", "tanh", "tanh"),
    out_func: str = "linear",
    optimizer: Union[str, Optimizer] = "Adam",
    optimizer_kwargs: Dict[str, Any] = dict(),
    compile_kwargs: Dict[str, Any] = dict(),
    **kwargs,
) -> keras.models.Sequential:
    """
    Builds a customized keras neural network auto-encoder based on a config dict

    Parameters
    ----------
    n_features: int
        Number of features the dataset X will contain.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.
    encoding_dim: tuple
        Tuple of numbers with the number of neurons in the encoding part.
    decoding_dim: tuple
        Tuple of numbers with the number of neurons in the decoding part.
    encoding_func: tuple
        Activation functions for the encoder part.
    decoding_func: tuple
        Activation functions for the decoder part.
    out_func: str
        Activation function for the output layer
    optimizer: Union[str, Optimizer]
        If str then the name of the optimizer must be provided (e.x. "Adam").
        The arguments of the optimizer can be supplied in optimize_kwargs.
        If a Keras optimizer call the instance of the respective
        class (e.x. Adam(lr=0.01,beta_1=0.9, beta_2=0.999)).  If no arguments are
        provided Keras default values will be set.
    optimizer_kwargs: Dict[str, Any]
        The arguments for the chosen optimizer. If not provided Keras'
        default values will be used.
    compile_kwargs: Dict[str, Any]
        Parameters to pass to ``keras.Model.compile``.

    Returns
    -------
    keras.models.Sequential

    """

    input_dim = n_features
    n_features_out = n_features_out or n_features

    check_dim_func_len("encoding", encoding_dim, encoding_func)
    check_dim_func_len("decoding", decoding_dim, decoding_func)

    model = KerasSequential()

    # Add encoding layers
    for i, (units, activation) in enumerate(zip(encoding_dim, encoding_func)):

        args = {"units": units, "activation": activation}

        if i == 0:
            args["input_dim"] = input_dim
        else:
            args["activity_regularizer"] = regularizers.l1(10e-5)

        model.add(Dense(**args))

    # Add decoding layers
    for i, (units, activation) in enumerate(zip(decoding_dim, decoding_func)):
        model.add(Dense(units=units, activation=activation))

    # Instantiate optimizer with kwargs
    if isinstance(optimizer, str):
        Optim = getattr(keras.optimizers, optimizer)
        optimizer = Optim(**optimizer_kwargs)

    # Final output layer
    model.add(Dense(n_features_out, activation=out_func))

    # Set some pre-determined default compile kwargs.
    compile_kwargs.update({"optimizer": optimizer})
    compile_kwargs.setdefault("loss", "mean_squared_error")
    compile_kwargs.setdefault("metrics", ["accuracy"])

    model.compile(**compile_kwargs)
    return model


@register_model_builder(type="KerasAutoEncoder")
def feedforward_symmetric(
    n_features: int,
    n_features_out: int = None,
    dims: Tuple[int, ...] = (256, 128, 64),
    funcs: Tuple[str, ...] = ("tanh", "tanh", "tanh"),
    optimizer: Union[str, Optimizer] = "Adam",
    optimizer_kwargs: Dict[str, Any] = dict(),
    compile_kwargs: Dict[str, Any] = dict(),
    **kwargs,
) -> keras.models.Sequential:
    """
    Builds a symmetrical feedforward model

    Parameters
    ----------
    n_features: int
         Number of input and output neurons.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.
    dim: List[int]
         Number of neurons per layers for the encoder, reversed for the decoder.
         Must have len > 0.
    funcs: List[str]
        Activation functions for the internal layers
    optimizer: Union[str, Optimizer]
        If str then the name of the optimizer must be provided (e.x. "Adam").
        The arguments of the optimizer can be supplied in optimization_kwargs.
        If a Keras optimizer call the instance of the respective
        class (e.x. Adam(lr=0.01,beta_1=0.9, beta_2=0.999)).  If no arguments are
        provided Keras default values will be set.
    optimizer_kwargs: Dict[str, Any]
        The arguments for the chosen optimizer. If not provided Keras'
        default values will be used.
    compile_kwargs: Dict[str, Any]
        Parameters to pass to ``keras.Model.compile``.

    Returns
    -------
    keras.models.Sequential

    """
    if len(dims) == 0:
        raise ValueError("Parameter dims must have len > 0")
    return feedforward_model(
        n_features,
        n_features_out,
        encoding_dim=dims,
        decoding_dim=dims[::-1],
        encoding_func=funcs,
        decoding_func=funcs[::-1],
        optimizer=optimizer,
        optimizer_kwargs=optimizer_kwargs,
        compile_kwargs=compile_kwargs,
        **kwargs,
    )


@register_model_builder(type="KerasAutoEncoder")
def feedforward_hourglass(
    n_features: int,
    n_features_out: int = None,
    encoding_layers: int = 3,
    compression_factor: float = 0.5,
    func: str = "tanh",
    optimizer: Union[str, Optimizer] = "Adam",
    optimizer_kwargs: Dict[str, Any] = dict(),
    compile_kwargs: Dict[str, Any] = dict(),
    **kwargs,
) -> keras.models.Sequential:
    """
    Builds an hourglass shaped neural network, with decreasing number of neurons
    as one gets deeper into the encoder network and increasing number
    of neurons as one gets out of the decoder network.

    Parameters
    ----------
    n_features: int
        Number of input and output neurons.
    n_features_out: Optional[int]
        Number of features the model will output, default to ``n_features``.
    encoding_layers: int
        Number of layers from the input layer (exclusive) to the
        narrowest layer (inclusive). Must be > 0. The total nr of layers
        including input and output layer will be 2*encoding_layers + 1.
    compression_factor: float
        How small the smallest layer is as a ratio of n_features
        (smallest layer is rounded up to nearest integer). Must satisfy
        0 <= compression_factor <= 1.
    func: str
        Activation function for the internal layers
    optimizer: Union[str, Optimizer]
        If str then the name of the optimizer must be provided (e.x. "Adam").
        The arguments of the optimizer can be supplied in optimization_kwargs.
        If a Keras optimizer call the instance of the respective
        class (e.x. Adam(lr=0.01,beta_1=0.9, beta_2=0.999)).  If no arguments are
        provided Keras default values will be set.
    optimizer_kwargs: Dict[str, Any]
        The arguments for the chosen optimizer. If not provided Keras'
        default values will be used.
    compile_kwargs: Dict[str, Any]
        Parameters to pass to ``keras.Model.compile``.

    Notes
    -----
    The resulting model will look like this when n_features = 10, encoding_layers= 3,
    and compression_factor = 0.3::

                * * * * * * * * * *
                  * * * * * * * *
                     * * * * *
                       * * *
                       * * *
                     * * * * *
                  * * * * * * * *
                * * * * * * * * * *


    Returns
    -------
    keras.models.Sequential

    Examples
    --------
    >>> model = feedforward_hourglass(10)
    >>> len(model.layers)
    7
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [8, 7, 5, 5, 7, 8, 10]
    >>> model = feedforward_hourglass(5)
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [4, 4, 3, 3, 4, 4, 5]
    >>> model = feedforward_hourglass(10, compression_factor=0.2)
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [7, 5, 2, 2, 5, 7, 10]
    >>> model = feedforward_hourglass(10, encoding_layers=1)
    >>> [model.layers[i].units for i in range(len(model.layers))]
    [5, 5, 10]
    """
    dims = hourglass_calc_dims(compression_factor, encoding_layers, n_features)

    return feedforward_symmetric(
        n_features,
        n_features_out,
        dims=dims,
        funcs=tuple([func] * len(dims)),
        optimizer=optimizer,
        optimizer_kwargs=optimizer_kwargs,
        compile_kwargs=compile_kwargs,
        **kwargs,
    )



# -*- coding: utf-8 -*-

import abc
import logging
import io
from pprint import pprint
from typing import Union, Callable, Dict, Any, Optional
from abc import ABCMeta

import h5py
import tensorflow.keras.models
from tensorflow.keras.models import load_model, save_model
from tensorflow.keras.preprocessing.sequence import pad_sequences, TimeseriesGenerator
from tensorflow.keras.wrappers.scikit_learn import KerasRegressor as BaseWrapper
import numpy as np
import pandas as pd

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.metrics import explained_variance_score
from sklearn.exceptions import NotFittedError

from gordo import serializer
from gordo.machine.model.base import GordoBase

# This is required to run `register_model_builder` against registered factories
from gordo.machine.model.factories import *  # pragma: no flakes

from gordo.machine.model.register import register_model_builder


logger = logging.getLogger(__name__)


class KerasBaseEstimator(BaseWrapper, GordoBase, BaseEstimator):
    def __init__(
        self,
        kind: Union[
            str, Callable[[int, Dict[str, Any]], tensorflow.keras.models.Model]
        ],
        **kwargs,
    ) -> None:
        """
        Initialized a Scikit-Learn API compatitble Keras model with a pre-registered
        function or a builder function
        directly.

        Parameters
        ----------
        kind: Union[callable, str]
            The structure of the model to build. As designated by any registered builder
            functions, registered with
            `gordo_compontents.model.register.register_model_builder`.
            Alternatively, one may pass a builder function directly to this argument.
            Such a function should accept `n_features` as it's first argument, and pass
            any additional parameters to `**kwargs`

        kwargs: dict
            Any additional args which are passed to the factory
            building function and/or any additional args to be passed
            to Keras' fit() method
        """
        self.build_fn = None
        self.kwargs = kwargs

        class_name = self.__class__.__name__

        if callable(kind):
            register_model_builder(type=class_name)(kind)
            self.kind = kind.__name__
        else:
            if kind not in register_model_builder.factories[class_name]:
                raise ValueError(
                    f"kind: {kind} is not an available model for type: {class_name}!"
                )
            self.kind = kind

    @property
    def sk_params(self):
        """
        Parameters used for scikit learn kwargs"""
        return self.kwargs

    def __getstate__(self):

        state = self.__dict__.copy()

        if hasattr(self, "model") and self.model is not None:
            buf = io.BytesIO()
            with h5py.File(buf, compression="lzf", mode="w") as h5:
                save_model(self.model, h5, overwrite=True, save_format="h5")
                buf.seek(0)
                state["model"] = buf
            if hasattr(self.model, "history"):
                from tensorflow.python.keras.callbacks import History

                history = History()
                history.history = self.model.history.history
                history.params = self.model.history.params
                history.epoch = self.model.history.epoch
                state["history"] = history
        return state

    def __setstate__(self, state):
        if "model" in state:
            with h5py.File(state["model"], compression="lzf", mode="r") as h5:
                state["model"] = load_model(h5, compile=False)
            if "history" in state:
                state["model"].__dict__["history"] = state.pop("history")
        self.__dict__ = state
        return self

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """
        Fit the model to X given y.

        Parameters
        ----------
        X: np.ndarray
            numpy array or pandas dataframe
        y: np.ndarray
            numpy array or pandas dataframe
        sample_weight: np.ndarray
            array like - weight to assign to samples
        kwargs
            Any additional kwargs to supply to keras fit method.

        Returns
        -------
        self
            'KerasAutoEncoder'
        """

        X = X.values if hasattr(X, "values") else X
        y = y.values if hasattr(y, "values") else y

        # Reshape y if needed, and set n features of target
        if y.ndim == 1:
            y = y.reshape(-1, 1)
        self.kwargs.update({"n_features_out": y.shape[1]})

        logger.debug(f"Fitting to data of length: {len(X)}")
        if len(X.shape) == 2:
            self.kwargs.update({"n_features": X.shape[1]})
        # for LSTM based models
        if len(X.shape) == 3:
            self.kwargs.update({"n_features": X.shape[2]})
        kwargs.setdefault("verbose", 0)
        super().fit(X, y, sample_weight=None, **kwargs)
        return self

    def predict(self, X: np.ndarray, **kwargs) -> np.ndarray:
        """

        Parameters
        ----------
        X: np.ndarray
            Input data
        kwargs: dict
            kwargs which are passed to Kera's ``predict`` method


        Returns
        -------
        results:
            np.ndarray
        """
        return self.model.predict(X, **kwargs)

    def get_params(self, **params):
        """
        Gets the parameters for this estimator

        Parameters
        ----------
        params
            ignored (exists for API compatibility).

        Returns
        -------
        Dict[str, Any]
            Parameters used in this estimator
        """
        params = super().get_params(**params)
        params.pop("build_fn", None)
        params.update({"kind": self.kind})
        params.update(self.kwargs)
        return params

    def __call__(self):
        build_fn = register_model_builder.factories[self.__class__.__name__][self.kind]
        return build_fn(**self.sk_params)

    def get_metadata(self):
        """
        Get metadata for the KerasBaseEstimator.
        Includes a dictionary with key "history". The key's value is a a dictionary
        with a key "params" pointing another dictionary with various parameters.
        The metrics are defined in the params dictionary under "metrics".
        For each of the metrics there is a key who's value is a list of values for this
        metric per epoch.

        Returns
        -------
        Dict
            Metadata dictionary, including a history object if present
        """
        if (
            hasattr(self, "model")
            and hasattr(self.model, "history")
            and self.model.history
        ):
            history = self.model.history.history
            history["params"] = self.model.history.params
            return {"history": history}
        else:
            return {}


class KerasAutoEncoder(KerasBaseEstimator, TransformerMixin):
    """
    Subclass of the KerasBaseEstimator to allow fitting to just X without requiring y.
    """

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        """
        Returns the explained variance score between auto encoder's input vs output

        Parameters
        ----------
        X: Union[np.ndarray, pd.DataFrame]
            Input data to the model
        y: Union[np.ndarray, pd.DataFrame]
            Target
        sample_weight: Optional[np.ndarray]
            sample weights

        Returns
        -------
        score: float
            Returns the explained variance score
        """
        if not hasattr(self, "model"):
            raise NotFittedError(
                f"This {self.__class__.__name__} has not been fitted yet."
            )

        out = self.model.predict(X)

        return explained_variance_score(y, out)


class KerasRawModelRegressor(KerasAutoEncoder):
    """
    Create a scikit-learn like model with an underlying tensorflow.keras model
    from a raw config.

    Examples
    --------
    >>> import yaml
    >>> import numpy as np
    >>> config_str = '''
    ...   # Arguments to the .compile() method
    ...   compile:
    ...     loss: mse
    ...     optimizer: adam
    ...
    ...   # The architecture of the model itself.
    ...   spec:
    ...     tensorflow.keras.models.Sequential:
    ...       layers:
    ...         - tensorflow.keras.layers.Dense:
    ...             units: 4
    ...         - tensorflow.keras.layers.Dense:
    ...             units: 1
    ... '''
    >>> config = yaml.safe_load(config_str)
    >>> model = KerasRawModelRegressor(kind=config)
    >>>
    >>> X, y = np.random.random((10, 4)), np.random.random((10, 1))
    >>> model.fit(X, y)
    KerasRawModelRegressor(kind: {'compile': {'loss': 'mse', 'optimizer': 'adam'},
     'spec': {'tensorflow.keras.models.Sequential': {'layers': [{'tensorflow.keras.layers.Dense': {'units': 4}},
                                                                {'tensorflow.keras.layers.Dense': {'units': 1}}]}}}
    )
    >>> out = model.predict(X)
    """

    def __init__(self, kind: dict, **kwargs):

        self.kind = kind  # type: ignore
        self.kwargs = kwargs

    def __repr__(self):
        stream = io.StringIO()
        pprint(self.kind, stream=stream)
        stream.seek(0)
        result = f"{self.__class__.__name__}(kind: {stream.read()})"
        return result

    def build_fn(self):
        """Build Keras model from specification"""
        _expected_keys = ("spec", "compile")
        if not all(k in self.kind for k in _expected_keys):
            raise ValueError(
                f"Expected spec to have keys: {_expected_keys}, but found {self.kind.keys()}"
            )
        logger.debug(f"Building model from spec: {self.kind}")

        model = serializer.from_definition(self.kind["spec"])

        # Load any compile kwargs as well, such as compile.optimizer which may map to class obj
        kwargs = serializer.from_definition(self.kind["compile"])

        model.compile(**kwargs)
        return model


class KerasLSTMBaseEstimator(KerasBaseEstimator, TransformerMixin, metaclass=ABCMeta):
    """
    Abstract Base Class to allow to train a many-one LSTM autoencoder and an LSTM
    1 step forecast
    """

    def __init__(
        self,
        kind: Union[Callable, str],
        lookback_window: int = 1,
        batch_size: int = 32,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        kind: Union[Callable, str]
            The structure of the model to build. As designated by any registered builder
            functions, registered with
            `gordo.machine.model.register.register_model_builder`.
            Alternatively, one may pass a builder function directly to this argument.
            Such a function should accept `n_features` as it's first argument, and pass
            any additional parameters to `**kwargs`.
        lookback_window: int
            Number of timestamps (lags) used to train the model.
        batch_size: int
            Number of training examples used in one epoch.
        epochs: int
            Number of epochs to train the model. An epoch is an iteration over the
            entire data provided.
        verbose: int
            Verbosity mode. Possible values are 0, 1, or 2 where 0 = silent,
            1 = progress bar, 2 = one line per epoch.
        kwargs: dict
            Any arguments which are passed to the factory building function and/or any
            additional args to be passed to the intermediate fit method.
        """
        self.lookback_window = lookback_window
        self.batch_size = batch_size
        kwargs["lookback_window"] = lookback_window
        kwargs["kind"] = kind
        kwargs["batch_size"] = batch_size

        # fit_generator_params is a set of strings with the keyword arguments of
        # Keras fit_generator method (excluding "shuffle" as this will be hardcoded).
        # This will be used in the fit method of the respective subclasses to match
        # the kwargs supplied when instantiating the subclass. The matched kwargs
        # will override the default kwargs of Keras fit_generator method when
        # training the model. Note: The decorator
        # "@interfaces.legacy_generator_methods_support" to Keras' fit_generator
        # method does not forward any arguments to the inspect module
        self.fit_generator_params = {
            "steps_per_epoch",
            "epochs",
            "verbose",
            "callbacks",
            "validation_data",
            "validation_steps",
            "validation_freq",
            "class_weight",
            "max_queue_size",
            "workers",
            "use_multiprocessing",
            "initial_epoch",
        }
        super().__init__(**kwargs)

    @abc.abstractproperty
    def lookahead(self) -> int:
        """Steps ahead in y the model should target"""
        ...

    def get_metadata(self):
        """
        Add number of forecast steps to metadata

        Returns
        -------
        metadata: dict
            Metadata dictionary, including forecast steps.
        """
        metadata = super().get_metadata()
        metadata.update({"forecast_steps": self.lookahead})
        return metadata

    def _validate_and_fix_size_of_X(self, X):
        if X.ndim == 1:
            logger.info(
                f"Reshaping X from an array to an matrix of shape {(len(X), 1)}"
            )
            X = X.reshape(len(X), 1)

        if self.lookback_window >= X.shape[0]:
            raise ValueError(
                "For KerasLSTMForecast lookback_window must be < size of X"
            )
        return X

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs) -> "KerasLSTMForecast":

        """
        This fits a one step forecast LSTM architecture.

        Parameters
        ----------
        X: np.ndarray
           2D numpy array of dimension n_samples x n_features. Input data to train.
        y: np.ndarray
           2D numpy array representing the target
        kwargs: dict
            Any additional args to be passed to Keras `fit_generator` method.

        Returns
        -------
        class:
            KerasLSTMForecast

        """

        X = X.values if isinstance(X, pd.DataFrame) else X
        y = y.values if isinstance(y, pd.DataFrame) else y

        X = self._validate_and_fix_size_of_X(X)

        # We call super.fit on a single sample (notice the batch_size=1) to initiate the
        # model using the scikit-learn wrapper.
        tsg = create_keras_timeseriesgenerator(
            X=X[
                : self.lookahead + self.lookback_window
            ],  # We only need a bit of the data
            y=y[: self.lookahead + self.lookback_window],
            batch_size=1,
            lookback_window=self.lookback_window,
            lookahead=self.lookahead,
        )

        primer_x, primer_y = tsg[0]

        super().fit(X=primer_x, y=primer_y, epochs=1, verbose=0)

        tsg = create_keras_timeseriesgenerator(
            X=X,
            y=y,
            batch_size=self.batch_size,
            lookback_window=self.lookback_window,
            lookahead=self.lookahead,
        )

        gen_kwargs = {
            k: v
            for k, v in {**self.kwargs, **kwargs}.items()
            if k in self.fit_generator_params
        }

        # shuffle is set to False since we are dealing with time series data and
        # so training data will not be shuffled before each epoch.
        self.model.fit_generator(tsg, shuffle=False, **gen_kwargs)
        return self

    def predict(self, X: np.ndarray, **kwargs) -> np.ndarray:
        """
        Parameters
        ----------
         X: np.ndarray
            Data to predict/transform. 2D numpy array of dimension `n_samples x
            n_features` where `n_samples` must be > lookback_window.

        Returns
        -------
        results: np.ndarray
                 2D numpy array of dimension `(n_samples - lookback_window) x
                 2*n_features`.  The first half of the array `(results[:,
                 :n_features])` corresponds to X offset by `lookback_window+1` (i.e.,
                 `X[lookback_window:,:]`) whereas the second half corresponds to the
                 predicted values of `X[lookback_window:,:]`.


        Example
        -------
        >>> import numpy as np
        >>> from gordo.machine.model.factories.lstm_autoencoder import lstm_model
        >>> from gordo.machine.model.models import KerasLSTMForecast
        >>> #Define train/test data
        >>> X_train = np.array([[1, 1], [2, 3], [0.5, 0.6], [0.3, 1], [0.6, 0.7]])
        >>> X_test = np.array([[2, 3], [1, 1], [0.1, 1], [0.5, 2]])
        >>> #Initiate model, fit and transform
        >>> lstm_ae = KerasLSTMForecast(kind="lstm_model",
        ...                             lookback_window=2,
        ...                             verbose=0)
        >>> model_fit = lstm_ae.fit(X_train, y=X_train.copy())
        >>> model_transform = lstm_ae.predict(X_test)
        >>> model_transform.shape
        (2, 2)
        """
        X = X.values if isinstance(X, pd.DataFrame) else X

        X = self._validate_and_fix_size_of_X(X)
        tsg = create_keras_timeseriesgenerator(
            X=X,
            y=X,
            batch_size=10000,
            lookback_window=self.lookback_window,
            lookahead=self.lookahead,
        )
        return self.model.predict_generator(tsg)

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        """
        Returns the explained variance score between 1 step forecasted input and true
        input at next time step (note: for LSTM X is offset by `lookback_window`).

        Parameters
        ----------
        X: Union[np.ndarray, pd.DataFrame]
            Input data to the model.
        y: Union[np.ndarray, pd.DataFrame]
            Target
        sample_weight: Optional[np.ndarray]
            Sample weights

        Returns
        -------
        score: float
            Returns the explained variance score.
        """
        if not hasattr(self, "model"):
            raise NotFittedError(
                f"This {self.__class__.__name__} has not been fitted yet."
            )

        out = self.predict(X)

        # Limit X samples to match the offset causes by LSTM lookback window
        # ie, if look back window is 5, 'out' will be 5 rows less than X by now
        return explained_variance_score(y[-len(out) :], out)


class KerasLSTMForecast(KerasLSTMBaseEstimator):
    @property
    def lookahead(self) -> int:
        return 1


class KerasLSTMAutoEncoder(KerasLSTMBaseEstimator):
    @property
    def lookahead(self) -> int:
        return 0


def create_keras_timeseriesgenerator(
    X: np.ndarray,
    y: Optional[np.ndarray],
    batch_size: int,
    lookback_window: int,
    lookahead: int,
) -> tensorflow.keras.preprocessing.sequence.TimeseriesGenerator:
    """
    Provides a `keras.preprocessing.sequence.TimeseriesGenerator` for use with
    LSTM's, but with the added ability to specify the lookahead of the target in y.

    If lookahead==0 then the generated samples in X will have as their last element
    the same as the corresponding Y. If lookahead is 1 then the values in Y is shifted
    so it is one step in the future compared to the last value in the samples in X,
    and similar for larger values.


    Parameters
    ----------
    X: np.ndarray
        2d array of values, each row being one sample.
    y: Optional[np.ndarray]
        array representing the target.
    batch_size: int
        How big should the generated batches be?
    lookback_window: int
        How far back should each sample see. 1 means that it contains a single
        measurement
    lookahead: int
        How much is Y shifted relative to X

    Returns
    -------
    TimeseriesGenerator
        3d matrix with a list of batchX-batchY pairs, where batchX is a batch of
        X-values, and correspondingly for batchY. A batch consist of `batch_size` nr
        of pairs of samples (or y-values), and each sample is a list of length
        `lookback_window`.

    Examples
    -------
    >>> import numpy as np
    >>> X, y = np.random.rand(100,2), np.random.rand(100, 2)
    >>> gen = create_keras_timeseriesgenerator(X, y,
    ...                                        batch_size=10,
    ...                                        lookback_window=20,
    ...                                        lookahead=0)
    >>> len(gen) # 9 = (100-20+1)/10
    9
    >>> len(gen[0]) # batchX and batchY
    2
    >>> len(gen[0][0]) # batch_size=10
    10
    >>> len(gen[0][0][0]) # a single sample, lookback_window = 20,
    20
    >>> len(gen[0][0][0][0]) # n_features = 2
    2
    """
    new_length = len(X) + 1 - lookahead
    kwargs: Dict[str, Any] = dict(length=lookback_window, batch_size=batch_size)
    if lookahead == 1:
        kwargs.update(dict(data=X, targets=y))

    elif lookahead >= 0:

        pad_kw = dict(maxlen=new_length, dtype=X.dtype)

        if lookahead == 0:
            kwargs["data"] = pad_sequences([X], padding="post", **pad_kw)[0]
            kwargs["targets"] = pad_sequences([y], padding="pre", **pad_kw)[0]

        elif lookahead > 1:
            kwargs["data"] = pad_sequences(
                [X], padding="post", truncating="post", **pad_kw
            )[0]
            kwargs["targets"] = pad_sequences(
                [y], padding="pre", truncating="pre", **pad_kw
            )[0]
    else:
        raise ValueError(f"Value of `lookahead` can not be negative, is {lookahead}")

    return TimeseriesGenerator(**kwargs)



# -*- coding: utf-8 -*-

import logging
import yaml
import json
import pydoc

import pytest

from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.decomposition import TruncatedSVD, PCA
from sklearn.preprocessing import MinMaxScaler, FunctionTransformer

from gordo.machine.model.models import KerasAutoEncoder
from gordo.serializer import into_definition, from_definition
from gordo.machine.model.register import register_model_builder

logger = logging.getLogger(__name__)


@pytest.fixture
def variations_of_same_pipeline():
    return [
        # Normal
        Pipeline(
            [
                ("pca1", PCA(n_components=2)),
                (
                    "fu",
                    FeatureUnion(
                        [
                            ("pca2", PCA(n_components=3)),
                            (
                                "pipe",
                                Pipeline(
                                    [
                                        ("minmax", MinMaxScaler()),
                                        ("truncsvd", TruncatedSVD(n_components=2)),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
            ]
        ),
        # MinMax initialized (wrongly) with a list
        Pipeline(
            [
                ("pca1", PCA(n_components=2)),
                (
                    "fu",
                    FeatureUnion(
                        [
                            ("pca2", PCA(n_components=3)),
                            (
                                "pipe",
                                Pipeline(
                                    [
                                        ("minmax", MinMaxScaler([0, 1])),
                                        ("truncsvd", TruncatedSVD(n_components=2)),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
            ]
        ),
        # MinMax initialized with tuple
        Pipeline(
            [
                ("pca1", PCA(n_components=2)),
                (
                    "fu",
                    FeatureUnion(
                        [
                            ("pca2", PCA(n_components=3)),
                            (
                                "pipe",
                                Pipeline(
                                    [
                                        ("minmax", MinMaxScaler((0, 1))),
                                        ("truncsvd", TruncatedSVD(n_components=2)),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
            ]
        ),
        # First pipeline without explicit steps param, other with.
        Pipeline(
            [
                ("pca1", PCA(n_components=2)),
                (
                    "fu",
                    FeatureUnion(
                        [
                            ("pca2", PCA(n_components=3)),
                            (
                                "pipe",
                                Pipeline(
                                    steps=[
                                        ("minmax", MinMaxScaler((0, 1))),
                                        ("truncsvd", TruncatedSVD(n_components=2)),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
            ]
        ),
    ]


def test_into_definition(variations_of_same_pipeline):

    expected_definition = """
        sklearn.pipeline.Pipeline:
            memory: null
            steps:
                - sklearn.decomposition._pca.PCA:
                    copy: true
                    iterated_power: auto
                    n_components: 2
                    random_state: null
                    svd_solver: auto
                    tol: 0.0
                    whiten: false
                - sklearn.pipeline.FeatureUnion:
                    n_jobs: null
                    transformer_list:
                    - sklearn.decomposition._pca.PCA:
                        copy: true
                        iterated_power: auto
                        n_components: 3
                        random_state: null
                        svd_solver: auto
                        tol: 0.0
                        whiten: false
                    - sklearn.pipeline.Pipeline:
                        memory: null
                        steps:
                        - sklearn.preprocessing._data.MinMaxScaler:
                            copy: true
                            feature_range:
                              - 0
                              - 1
                        - sklearn.decomposition._truncated_svd.TruncatedSVD:
                            algorithm: randomized
                            n_components: 2
                            n_iter: 5
                            random_state: null
                            tol: 0.0
                        verbose: false
                    transformer_weights: null
                    verbose: false
                - gordo.machine.model.models.KerasAutoEncoder:
                    kind: feedforward_hourglass
            verbose: false
        """

    expected_definition = yaml.safe_load(expected_definition)

    for pipe in variations_of_same_pipeline:

        definition = into_definition(from_definition(into_definition(pipe)))

        assert json.dumps(definition) == json.dumps(
            expected_definition
        ), f"Failed output:\n{definition}\nExpected:----------------\n{expected_definition}"


def test_into_from():
    """
    Pass Pipeline into definition, and then from that definition
    """
    from gordo.machine.model.transformer_funcs.general import multiply_by

    factories = register_model_builder.factories
    for model in factories.keys():

        for model_kind in factories[model].keys():
            pipe = Pipeline(
                [
                    ("step_0", PCA(n_components=2)),
                    (
                        "step_1",
                        FeatureUnion(
                            [
                                ("step_0", PCA(n_components=3)),
                                (
                                    "step_1",
                                    Pipeline(
                                        steps=[
                                            ("step_0", MinMaxScaler((0, 1))),
                                            ("step_1", TruncatedSVD(n_components=2)),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    ),
                    (
                        "step_2",
                        FunctionTransformer(func=multiply_by, kw_args={"factor": 1}),
                    ),
                    (
                        "step_3",
                        pydoc.locate(f"gordo.machine.model.models.{model}")(
                            kind=model_kind
                        ),
                    ),
                ]
            )

            from_definition(into_definition(pipe))


def test_from_into():
    """
    Create pipeline from definition, and create from that definition
    """
    factories = register_model_builder.factories
    for model in factories.keys():
        for model_kind in factories[model].keys():
            definition = f"""
                sklearn.pipeline.Pipeline:
                    steps:
                        - sklearn.decomposition.pca.PCA:
                            n_components: 2
                            copy: true
                            whiten: false
                            svd_solver: auto
                            tol: 0.0
                            iterated_power: auto
                            random_state:
                        - sklearn.preprocessing._function_transformer.FunctionTransformer:
                            func: gordo.machine.model.transformer_funcs.general.multiply_by
                            kw_args:
                                factor: 1
                            inverse_func: gordo.machine.model.transformer_funcs.general.multiply_by
                            inv_kw_args:
                                factor: 1
                        - sklearn.pipeline.FeatureUnion:
                            transformer_list:
                            - sklearn.decomposition.pca.PCA:
                                n_components: 3
                                copy: true
                                whiten: false
                                svd_solver: auto
                                tol: 0.0
                                iterated_power: auto
                                random_state:
                            - sklearn.pipeline.Pipeline:
                                steps:
                                - sklearn.preprocessing.data.MinMaxScaler:
                                    feature_range:
                                    - 0
                                    - 1
                                    copy: true
                                - sklearn.decomposition.truncated_svd.TruncatedSVD:
                                    n_components: 2
                                    algorithm: randomized
                                    n_iter: 5
                                    random_state:
                                    tol: 0.0
                                memory:
                                verbose: false
                            n_jobs: 1
                            transformer_weights:
                            verbose: false
                        - gordo.machine.model.models.{model}:
                            kind: {model_kind}
                    memory:
                    verbose: false
                """
            definition = yaml.safe_load(definition)
            pipe = from_definition(definition)
            into_definition(pipe)


def test_captures_kwarg_to_init():
    """
    Our models allow kwargs which are put into the underlying keras model or to construct
    the underlying model.
    We want to ensure into defintion captures kwargs which are part of the model
    parameters but not part of the __init__ signature
    """
    ae = KerasAutoEncoder(kind="feedforward_hourglass", some_fancy_param="Howdy!")
    definition = into_definition(ae)
    parameters = definition[
        f"{KerasAutoEncoder.__module__}.{KerasAutoEncoder.__name__}"
    ]
    assert "some_fancy_param" in parameters
    assert parameters["some_fancy_param"] == "Howdy!"

    # And make sure we can init again
    KerasAutoEncoder(**parameters)



import os
import yaml
import dateutil.parser
import logging
import jinja2
import io

from typing import Union

logger = logging.getLogger(__name__)


def _docker_friendly_version(version):
    """
    Some untagged versions may have a '+' in which case is not a valid
    docker tag
    """
    return version.replace("+", "_")


def _valid_owner_ref(owner_reference_str: str):
    """
    Validates that the parameter can be loaded (as yaml) into a non-empty list of
    valid owner-references.

    A valid owner reference is a dict containing at least the keys 'uid', 'name',
    'kind', and 'apiVersion'. Raises `TypeError` if not, or returns
    the list of owner references if they seem valid.

    Parameters
    ----------
    owner_reference_str: str
        String representation of the list of owner-references, should be parsable as
        yaml/json

    Returns
    -------
    list[dict]
        The list of owner-references

    """
    owner_ref = yaml.safe_load(owner_reference_str)
    if not type(owner_ref) == list or len(owner_ref) < 1:
        raise TypeError("Owner-references must be a list with at least one element")
    for oref in owner_ref:
        if (
            "uid" not in oref
            or "name" not in oref
            or "kind" not in oref
            or "apiVersion" not in oref
        ):
            raise TypeError(
                "All elements in owner-references must contain a uid, name, kind, "
                "and apiVersion key "
            )
    return owner_ref


def _timestamp_constructor(_loader, node):
    parsed_date = dateutil.parser.isoparse(node.value)
    if parsed_date.tzinfo is None:
        raise ValueError(
            "Provide timezone to timestamp {}."
            " Example: for UTC timezone use {} or {} ".format(
                node.value, node.value + "Z", node.value + "+00:00"
            )
        )
    return parsed_date


def get_dict_from_yaml(config_file: Union[str, io.StringIO]) -> dict:
    """
    Read a config file or file like object of YAML into a dict
    """
    # We must override the default constructor for timestamp to ensure the result
    # has tzinfo. Yaml loader never adds tzinfo, but converts to UTC.
    yaml.FullLoader.add_constructor(
        tag="tag:yaml.org,2002:timestamp", constructor=_timestamp_constructor
    )
    if hasattr(config_file, "read"):
        yaml_content = yaml.load(config_file, Loader=yaml.FullLoader)
    else:
        try:
            path_to_config_file = os.path.abspath(config_file)  # type: ignore
            with open(path_to_config_file, "r") as yamlfile:  # type: ignore
                yaml_content = yaml.load(yamlfile, Loader=yaml.FullLoader)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Unable to find config file <{path_to_config_file}>"
            )
    # Handle multiple versions of workflow config structure
    if "spec" in yaml_content:
        yaml_content = yaml_content["spec"]["config"]

    return yaml_content


def load_workflow_template(workflow_template: str) -> jinja2.Template:
    """
    Loads the Jinja2 Template from a specified path

    Parameters
    ----------
    workflow_template: str
        Path to a workflow template

    Returns
    -------
    jinja2.Template
        Loaded but non-rendered jinja2 template for the workflow
    """
    path_to_workflow_template = os.path.abspath(workflow_template)
    template_dir = os.path.dirname(path_to_workflow_template)

    templateEnv = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_dir), undefined=jinja2.StrictUndefined
    )
    return templateEnv.get_template(os.path.basename(workflow_template))



# -*- coding: utf-8 -*-

import pytest

from gordo.builder import local_build
from gordo.machine import Machine


@pytest.mark.parametrize(
    "config",
    (
        """
    machines:
      - dataset:
          tags:
            - SOME-TAG1
            - SOME-TAG2
          target_tag_list:
            - SOME-TAG3
            - SOME-TAG4
          train_end_date: '2019-03-01T00:00:00+00:00'
          train_start_date: '2019-01-01T00:00:00+00:00'
          asset: asgb
          data_provider:
            type: RandomDataProvider
        metadata:
          information: Some sweet information about the model
        model:
          gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
              sklearn.pipeline.Pipeline:
                steps:
                - sklearn.decomposition.pca.PCA
                - sklearn.multioutput.MultiOutputRegressor:
                    estimator: sklearn.linear_model.base.LinearRegression
        name: crazy-sweet-name
    """,
        """
    machines:
      - dataset:
          tags:
            - SOME-TAG1
            - SOME-TAG2
          train_end_date: '2019-03-01T00:00:00+00:00'
          train_start_date: '2019-01-01T00:00:00+00:00'
          asset: asgb
          data_provider:
            type: RandomDataProvider
        name: crazy-sweet-name
    globals:
        model:
          gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
              sklearn.pipeline.Pipeline:
                steps:
                - sklearn.decomposition.pca.PCA
                - sklearn.multioutput.MultiOutputRegressor:
                    estimator: sklearn.linear_model.base.LinearRegression
    """,
    ),
)
def test_local_builder_valid_configs(config):
    models_n_metadata = list(local_build(config))
    assert len(models_n_metadata) == 1

    model_n_metadata = models_n_metadata.pop()
    assert isinstance(model_n_metadata, tuple)
    assert len(model_n_metadata) == 2

    model, machine = model_n_metadata
    assert hasattr(model, "fit")
    assert isinstance(machine, Machine)


@pytest.mark.parametrize(
    "config",
    (
        """
        machines:
          - dataset:
              tags:
                - SOME-TAG1
                - SOME-TAG2
              target_tag_list:
                - SOME-TAG3
                - SOME-TAG4
              train_end_date: '2019-03-01T00:00:00+00:00'
              train_start_date: '2019-01-01T00:00:00+00:00'
              asset: asgb
              data_provider:
                type: RandomDataProvider
            metadata:
              information: Some sweet information about the model
            ## OH NO, NO MODEL HERE, FAIL ##
            name: crazy-sweet-name
        """,
        """
        machines:
          - dataset:
              tags:
                - SOME-TAG1
                - SOME-TAG2
              train_end_date: '2019-03-01T00:00:00+00:00'
              train_start_date: '2019-01-01T00:00:00+00:00'
              asset: asgb
              data_provider:
                type: RandomDataProvider
            name: crazy-sweet-name-  ## <- Cannot end with ending hyphen
        globals:
            model:
              gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
                base_estimator:
                  sklearn.pipeline.Pipeline:
                    steps:
                    - sklearn.decomposition.pca.PCA
                    - sklearn.multioutput.MultiOutputRegressor:
                        estimator: sklearn.linear_model.base.LinearRegression
        """,
    ),
)
def test_local_builder_invalid_configs(config):
    with pytest.raises(Exception):
        list(local_build(config))



import unittest
from io import StringIO
from unittest import mock

from dateutil.parser import isoparse  # type: ignore

from gordo.machine.dataset.data_provider.iroc_reader import IrocReader, read_iroc_file
from gordo.machine.dataset.sensor_tag import normalize_sensor_tags
from gordo.machine.dataset.sensor_tag import SensorTag

IROC_HAPPY_TAG_LIST = [
    SensorTag("NINENINE.OPCIS::NNFCDPC01.AI1410J0", "NINENINE"),
    SensorTag("NINENINE.OPCIS::NNFCDPC01.AI1840C1J0", "NINENINE"),
    SensorTag("NINENINE.OPCIS::NNFCDPC01.AI1840E1J0", "NINENINE"),
]

HAPPY_FROM_TS = isoparse("2018-05-02T01:56:00+00:00")
HAPPY_TO_TS = isoparse("2018-05-03T01:56:00+00:00")

IROC_MANY_ASSETS_TAG_LIST = [
    "NINENINE.OPCIS::NNFCDPC01.AI1410J0",
    "NINENINE.OPCIS::NNFCDPC01.AI1840C1J0",
    "NINENINE.OPCIS::NNFCDPC01.AI1840E1J0",
    "UON_EF.OPCIS::LO006-B1H.PRCASXIN",
    "UON_EF.OPCIS::LO006-B1H.PRTUBXIN",
    "UON_EF.OPCIS::LO006-B1H_M1.PRSTAXIN",
    "UON_EF.OPCIS::LO006-B1H_M1.RTGASDIN",
]

IROC_NO_ASSET_TAG_LIST = [
    SensorTag("NOT.OPCIS::NNFCDPC01.AI1410J0", "NOT"),
    SensorTag("AN.OPCIS::NNFCDPC01.AI1840C1J0", "AN"),
    SensorTag("IROC.OPCIS::NNFCDPC01.AI1840E1J0", "IROC"),
    SensorTag("ASSET.OPCIS::LO006-B1H.PRCASXIN", "ASSET"),
]

# Functioning CSV for IROC. Has 6 lines, 5 different timestamps
# (2018-05-02T06:44:29.7830000Z occurs twice) and 3 different tags.
IROC_HAPPY_PATH_CSV = u"""tag,value,timestamp,status
NINENINE.OPCIS::NNFCDPC01.AI1410J0,5,2018-05-02T06:00:11.3860000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1410J0,76.86899,2018-05-02T06:44:29.7830000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840C1J0,-23.147645,2018-05-02T06:43:53.8490000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840C1J0,-10.518037,2018-05-02T06:44:29.9130000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840E1J0,48.92137,2018-05-02T06:43:59.7240000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840E1J0,-0.497645,2018-05-02T06:44:29.7830000Z,Analog Normal
                """

IROC_MANY_ASSETS_SENSOR_TAG_LIST = [
    SensorTag("NINENINE.OPCIS::NNFCDPC01.AI1410J0", "ninenine"),
    SensorTag("NINENINE.OPCIS::NNFCDPC01.AI1840C1J0", "ninenine"),
    SensorTag("NINENINE.OPCIS::NNFCDPC01.AI1840E1J0", "ninenine"),
    SensorTag("UON_EF.OPCIS::LO006-B1H.PRCASXIN", "uon_ef"),
    SensorTag("UON_EF.OPCIS::LO006-B1H.PRTUBXIN", "uon_ef"),
    SensorTag("UON_EF.OPCIS::LO006-B1H_M1.PRSTAXIN", "uon_ef"),
    SensorTag("UON_EF.OPCIS::LO006-B1H_M1.RTGASDIN", "uon_ef"),
]


def test_normalize_iroc_tags():
    normalized_tags = normalize_sensor_tags(IROC_MANY_ASSETS_TAG_LIST)
    assert normalized_tags == IROC_MANY_ASSETS_SENSOR_TAG_LIST


class IrocDataSourceTestCase(unittest.TestCase):
    def test_read_iroc_file_basic(self):
        """A basic happy-path testing of reading an iroc CSV with some values"""
        f = StringIO(IROC_HAPPY_PATH_CSV)
        res_df = read_iroc_file(
            file_obj=f,
            train_start_date=HAPPY_FROM_TS,
            train_end_date=HAPPY_TO_TS,
            tag_list=IROC_HAPPY_TAG_LIST,
        )
        for tag in IROC_HAPPY_TAG_LIST:
            self.assertIn(tag.name, res_df.columns)
        # We have one row per distinct timestamp in the input-csv
        self.assertEqual(5, len(res_df))

    def test_read_iroc_file_with_errors(self):
        """Reading a csv with some of the 'values' beeing strings instead of numbers
        leads to those lines being ignored completely"""
        f = StringIO(
            u"""tag,value,timestamp,status
NINENINE.OPCIS::NNFCDPC01.AI1410J0,5,2018-05-02T06:00:11.3860000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1410J0,76.86899,2018-05-02T06:44:29.7830000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840C1J0,lameness,2018-05-02T06:43:53.8490000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840C1J0,-10.518037,2018-05-02T06:44:29.9130000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840E1J0,wtf,2018-05-02T06:43:59.7240000Z,Analog Normal
NINENINE.OPCIS::NNFCDPC01.AI1840E1J0,-0.497645,2018-05-02T06:44:29.7830000Z,Analog Normal
                """
        )

        res_df = read_iroc_file(
            file_obj=f,
            train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
            train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
            tag_list=IROC_HAPPY_TAG_LIST,
        )
        for tag in IROC_HAPPY_TAG_LIST:
            self.assertIn(tag.name, res_df.columns)
        # Two of the lines have invalid values (non-numeric), so those are ignored
        # and there are only 3 distinct timestamps left.
        self.assertEqual(3, len(res_df))

    @mock.patch.object(
        IrocReader,
        "_fetch_all_iroc_files_from_paths",
        side_effect=lambda all_base_paths, train_start_date, train_end_date, tag_list: [
            read_iroc_file(
                file_obj=StringIO(IROC_HAPPY_PATH_CSV),
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag_list=tag_list,
            )
        ],
    )
    def test_load_series_no_data(self, _mocked_method):
        """load_series will raise ValueError if it does not find any tags"""
        iroc_reader = IrocReader(client=None, threads=1)
        with self.assertRaises(ValueError):
            list(
                iroc_reader.load_series(
                    train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
                    train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                    tag_list=[SensorTag("jalla", None)],  # Not a tag in the input
                )
            )

    def test_load_series_no_tag_list(self):
        """load_series will return an empty generator when called with no tags"""
        iroc_reader = IrocReader(client=None, threads=1)
        res = list(
            iroc_reader.load_series(
                train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
                train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                tag_list=[],
            )
        )
        self.assertEqual([], res)

    def test_can_handle_tag_ok(self):
        iroc_reader = IrocReader(client=None, threads=1)
        assert iroc_reader.can_handle_tag(SensorTag("UON_EF.xxx", "UON_EF"))

    def test_can_handle_tag_unknown_asset(self):
        iroc_reader = IrocReader(client=None, threads=1)
        assert not iroc_reader.can_handle_tag(SensorTag("UON_EF.xxx", "UNKNOWǸ_ASSET"))

    def test_can_handle_tag_no_asset(self):
        iroc_reader = IrocReader(client=None, threads=1)
        assert not iroc_reader.can_handle_tag(SensorTag("UON_EF.xxx", None))

    def test_load_series_many_assets(self):
        """load_series will return an empty generator when called with tags
        related to several assets"""
        iroc_reader = IrocReader(client=None, threads=1)
        with self.assertRaises(ValueError):
            list(
                iroc_reader.load_series(
                    train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
                    train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                    tag_list=IROC_MANY_ASSETS_SENSOR_TAG_LIST,  # Not a tag in the input
                )
            )

    def test_load_series_no_asset_found(self):
        """load_series will return an empty generator when called with tags
        that cannot be related to any asset"""
        iroc_reader = IrocReader(client=None, threads=1)
        with self.assertRaises(ValueError):
            list(
                iroc_reader.load_series(
                    train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
                    train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                    tag_list=IROC_NO_ASSET_TAG_LIST,  # Not a tag in the input
                )
            )

    def test_load_series_checks_date(self):
        """load_series will raise ValueError if train_end_date<train_start_date"""
        iroc_reader = IrocReader(client=None, threads=1)
        with self.assertRaises(ValueError):
            list(
                iroc_reader.load_series(
                    train_start_date=isoparse("2018-05-03T01:56:00+00:00"),
                    train_end_date=isoparse("2018-05-02T01:56:00+00:00"),
                    tag_list=[SensorTag("jalla", None)],  # Not a tag in the input
                )
            )

    @mock.patch.object(
        IrocReader,
        "_fetch_all_iroc_files_from_paths",
        side_effect=lambda all_base_paths, train_start_date, train_end_date, tag_list: [
            read_iroc_file(
                file_obj=StringIO(IROC_HAPPY_PATH_CSV),
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag_list=tag_list,
            )
        ],
    )
    def test_load_series_missing_columns_data(self, _mocked_method):
        """load_series will raise ValueError if there is a single tag it can not
        find"""
        iroc_reader = IrocReader(client=None, threads=1)
        with self.assertRaises(ValueError):
            list(
                iroc_reader.load_series(
                    train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
                    train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                    tag_list=IROC_HAPPY_TAG_LIST + [SensorTag("jalla", None)],
                    # "jalla" is not a tag
                )
            )

    @mock.patch.object(
        IrocReader,
        "_fetch_all_iroc_files_from_paths",
        side_effect=lambda all_base_paths, train_start_date, train_end_date, tag_list: [
            read_iroc_file(
                file_obj=StringIO(IROC_HAPPY_PATH_CSV),
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag_list=tag_list,
            )
        ],
    )
    def test_load_series_happy_path(self, _mocked_method):
        """Happy-path testing of load_dataframe"""
        iroc_reader = IrocReader(client=None, threads=1)
        res = list(
            iroc_reader.load_series(
                train_start_date=isoparse("2018-05-02T01:56:00+00:00"),
                train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                tag_list=IROC_HAPPY_TAG_LIST,
            )
        )
        # We get one dataframe per tag, so 3
        self.assertEqual(3, len(res))

    @mock.patch.object(
        IrocReader,
        "_fetch_all_iroc_files_from_paths",
        side_effect=lambda all_base_paths, train_start_date, train_end_date, tag_list: [
            read_iroc_file(
                file_obj=StringIO(IROC_HAPPY_PATH_CSV),
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag_list=tag_list,
            )
        ],
    )
    def test_load_series_happy_path_different_timezones(self, _mocked_method):
        """Happy-path testing of load_dataframe"""
        iroc_reader = IrocReader(client=None, threads=1)
        res = list(
            iroc_reader.load_series(
                train_start_date=isoparse("2018-05-02T01:56:00+02:00"),
                train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                tag_list=IROC_HAPPY_TAG_LIST,
            )
        )
        # We get one dataframe per tag, so 3
        self.assertEqual(3, len(res))

    def test_load_series_dry_run_raises(self):
        iroc_reader = IrocReader(client=None)

        with self.assertRaises(NotImplementedError):
            list(
                iroc_reader.load_series(
                    train_start_date=isoparse("2018-05-02T01:56:00+02:00"),
                    train_end_date=isoparse("2018-05-03T01:56:00+00:00"),
                    tag_list=IROC_HAPPY_TAG_LIST,
                    dry_run=True,
                )
            )
